
#include "PID.h"

/*Constructor (...)*********************************************************
*    The parameters specified here are those for for which we can't set up 
*    reliable defaults, so we need to have the user set them.
***************************************************************************/
void PID_Create(PID_t* ctx, float Kp, float Ki, float Kd, u16 sampleTime, u8 ControllerDirection)
{
	ctx->output = 0.0;
	ctx->input = 0.0;
	ctx->lastInput = 0.0;
	
	PID_SetOutputLimits(ctx,0, 255);	// default output limit corresponds to 
																		// the arduino pwm limits

	ctx->sampleTime = sampleTime;			// default Controller Sample Time is 0.1 seconds

	PID_SetControllerDirection(ctx,ControllerDirection);
	PID_SetTunings(ctx,Kp, Ki, Kd);
	
	PID_Initialize(ctx);
}


/* Compute() **********************************************************************
*     This, as they say, is where the magic happens.  this function should be called
*   at sample time.  
**********************************************************************************/ 
float PID_Compute(PID_t* ctx,float input, float setpoint)
{
	float error;
	float dInput;
	float output;
	
	ctx->input = input;
	
	/*Compute all the working error variables*/
	error = setpoint - input;
	ctx->ITerm += (ctx->ki * error);
	if(ctx->ITerm > ctx->outMax) {
		ctx->ITerm= ctx->outMax;
	} else if(ctx->ITerm < ctx->outMin) {
		ctx->ITerm= ctx->outMin;
	}
	
	dInput = (input - ctx->lastInput);

	/*Compute PID Output*/
	output = ctx->kp * error + ctx->ITerm - ctx->kd * dInput;
	
	if(output > ctx->outMax) {
		output = ctx->outMax;
	} else if(output < ctx->outMin) { 
		output = ctx->outMin;
	}
	ctx->output = output;
	
	/*Remember some variables for next time*/
	ctx->lastInput = input;

	return ctx->output;
}


/* SetTunings(...)*************************************************************
* This function allows the controller's dynamic performance to be adjusted. 
* it's called automatically from the constructor, but tunings can also
* be adjusted on the fly during normal operation
******************************************************************************/ 
void PID_SetTunings(PID_t* ctx, float Kp, float Ki, float Kd)
{
	float SampleTimeInSec;
	
	if (Kp<0 || Ki<0 || Kd<0) 
		return;

	SampleTimeInSec = ((float)ctx->sampleTime)/1000;  
	ctx->kp = Kp;
	ctx->ki = Ki * SampleTimeInSec;
	ctx->kd = Kd / SampleTimeInSec;

	if(ctx->controllerDirection == REVERSE) {
		ctx->kp = (0 - ctx->kp);
		ctx->ki = (0 - ctx->ki);
		ctx->kd = (0 - ctx->kd);
	}
}

/* SetSampleTime(...) *********************************************************
* sets the period, in Milliseconds, at which the calculation is performed	
******************************************************************************/
void PID_SetSampleTime(PID_t* ctx, int NewSampleTime)
{
	if (NewSampleTime > 0) {
		float ratio  = (float)NewSampleTime / (float)ctx->sampleTime;
		ctx->ki *= ratio;
		ctx->kd /= ratio;
		ctx->sampleTime = (unsigned long)NewSampleTime;
	}
}

/* SetOutputLimits(...)****************************************************
*     This function will be used far more often than SetInputLimits.  while
*  the input to the controller will generally be in the 0-1023 range (which is
*  the default already,)  the output will be a little different.  maybe they'll
*  be doing a time window and will need 0-8000 or something.  or maybe they'll
*  want to clamp it from 0-125.  who knows.  at any rate, that can all be done
*  here.
**************************************************************************/
void PID_SetOutputLimits(PID_t* ctx, float Min, float Max)
{
	if(Min >= Max) return;
	ctx->outMin = Min;
	ctx->outMax = Max;

	if(ctx->output > ctx->outMax) ctx->output = ctx->outMax;
	else if(ctx->output < ctx->outMin) ctx->output = ctx->outMin;
	
	if(ctx->ITerm > ctx->outMax) ctx->ITerm= ctx->outMax;
	else if(ctx->ITerm < ctx->outMin) ctx->ITerm= ctx->outMin;
}

/* Initialize()****************************************************************
*	does all the things that need to happen to ensure a bumpless transfer
*  from manual to automatic mode.
******************************************************************************/ 
void PID_Initialize(PID_t* ctx)
{
	ctx->ITerm = ctx->output;
	ctx->lastInput = ctx->input;
	if(ctx->ITerm > ctx->outMax) {
		ctx->ITerm = ctx->outMax;
	} else if(ctx->ITerm < ctx->outMin) {
		ctx->ITerm = ctx->outMin;
	}
}

/* SetControllerDirection(...)*************************************************
* The PID will either be connected to a DIRECT acting process (+Output leads 
* to +Input) or a REVERSE acting process(+Output leads to -Input.)  we need to
* know which one, because otherwise we may increase the output when we should
* be decreasing.  This is called from the constructor.
******************************************************************************/
void PID_SetControllerDirection(PID_t* ctx, int Direction)
{
	if(Direction !=ctx->controllerDirection)
	{
		ctx->kp = (0 - ctx->kp);
		ctx->ki = (0 - ctx->ki);
		ctx->kd = (0 - ctx->kd);
	}   
	ctx->controllerDirection = Direction;
}

int PID_GetDirection(PID_t* ctx)
{ 
	return ctx->controllerDirection;
}
