//-----------------------------------------------------------------------------
// SolderStation_v2.c
//  Implements a SolderStation control loop for the ZD9xx series solderstations, 
//  using a temperature PID controller -- This version uses float precision floats
//
//  (C) 2012-2013 Eduardo Jos Tagle <ejtagle@hotmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ------------------------------------------------------------------------
//																	
// Port usage:
//	P0.1 O Buzzer: 1=Off. Set as open drain
//	P0.2 O Solder On
//	P1.2 A ThermoCouple
//	P1.4 A ThermoCouple GND - Not used
//	P2.0 I PushButton (Active LO) - Unused
//	P2.1 I PushButton (Active LO) - Celcius/Farenheit
//	P2.2 I PushButton (Active LO) - Down
//	P2.3 I PushButton (Active LO) - Up
//	P3.4 IO HT DATA
//	P3.3 O  HT WR
//  P3.2 O  HT CS
//
// ------------------------------------------------------------------------
//
// Peripheral use:
//	TMR0: PWM Timebase
//  TMR1: UART Timebase
//	TMR2: ISR Timebase
//	UART
//	ADC: Temperature sensing
//
// ------------------------------------------------------------------------
//
// TODO:
//	-Fine tune the temperature PID gain parameters by capturing the step response
//   of the heater, and applying Ziegler-Nichols PID parameter tuning
//  -Send tip temperature through rs232 for logging purposes.
//  -Autoshutdown ... ?
//  -Maybe implement an autotuning temperature controller...
//
// ------------------------------------------------------------------------
//  Any improvements, feature additions, please, share! ... And send me a copy of them!
//  <ejtagle@hotmail.com>
//


//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <c8051f310.h>                 // SFR declarations
#include <intrins.h>
#include <stdio.h>

//-----------------------------------------------------------------------------
// 16-bit SFR Definitions for 'F31x
//-----------------------------------------------------------------------------

sfr16 DP       = 0x82;                 // data pointer
sfr16 TMR2RL   = 0xca;                 // Timer2 reload value
sfr16 TMR2     = 0xcc;                 // Timer2 counter
sfr16 TMR3     = 0x94;                 // Timer3 counter
sfr16 TMR3RL   = 0x92;                 // Timer3 reload value
sfr16 PCA0CP0  = 0xfb;                 // PCA0 Module 0 Capture/Compare
sfr16 PCA0CP1  = 0xe9;                 // PCA0 Module 1 Capture/Compare
sfr16 PCA0CP2  = 0xeb;                 // PCA0 Module 2 Capture/Compare
sfr16 PCA0CP3  = 0xed;                 // PCA0 Module 3 Capture/Compare
sfr16 PCA0CP4  = 0xfd;                 // PCA0 Module 4 Capture/Compare
sfr16 PCA0     = 0xf9;                 // PCA0 counter
sfr16 ADC0     = 0xbd;                 // ADC Data Word Register
sfr16 ADC0GT   = 0xc3;                 // ADC0 Greater-Than
sfr16 ADC0LT   = 0xc5;                 // ADC0 Less-Than

//-----------------------------------------------------------------------------
// Port defines
//-----------------------------------------------------------------------------

//  P0.1 O Buzzer
//	P0.2 O Solder On
//	P1.2 A ThermoCouple
//	P1.4 A ThermoCouple GND
//	P2.0 I PushButton (Active LO) - Unused
//	P2.1 I PushButton (Active LO) - Celcius/Farenheit
//	P2.2 I PushButton (Active LO) - Down
//	P2.3 I PushButton (Active LO) - Up
//	P3.4 IO HT DATA
//	P3.3 O  HT WR
//  P3.2 O  HT CS

sbit BUZZER 	= P0^1; //	P0.1 O Buzzer 
sbit SOLDERON	= P0^2; //  P0.2 O SolderOn
sbit PB_UNUSED	= P2^0;	//	P2.0 I PushButton (Active LO) - Unused
sbit PB_CFTOGGLE= P2^1;	//	P2.1 I PushButton (Active LO) - Celcius/Farenheit
sbit PB_DOWN	= P2^2;	//	P2.2 I PushButton (Active LO) - Down
sbit PB_UP		= P2^3;	//	P2.3 I PushButton (Active LO) - Up
sbit HTDAT 		= P3^4; //	P3.4 IO HT DATA
sbit HTWR  		= P3^3; //	P3.3 O  HT WR
sbit HTCS  		= P3^2; //  P3.2 O  HT CS

//-----------------------------------------------------------------------------
// Global CONSTANTS
//-----------------------------------------------------------------------------

#define SYSCLK      24500000           // SYSCLK frequency in Hz
#define BAUDRATE        9600           // Baud rate of UART in bps
#define TIMER2_RATE     100            // Timer 2 overflow rate in Hz

#include "mytypes.h"
#include "PID.h"
#include "PID_AutoTune.h"

// PID Status and Autotuning
PID_t PID_State;

// PID Autotuning

// PID configuration
typedef struct { 
	float Kp; 			// Kp
	float Ki; 			// Ki
	float Kd; 			// Kd
} pid_cfg;

// System
struct sys_cfg {
	float setpoint;		// Temperature setpoint

	pid_cfg pid;		// PID configuration

	u8 use_fahrenheit;	// Si debemos usar dF or not
	u8 invalid;			// Invalid. (0=valid)
	u8 cksum;				// Checksum
};

//-----------------------------------------------------------------------------
// Global VARIABLES
//-----------------------------------------------------------------------------

// The nonvolatile configuration stored in config space... We have 2 copies. A master and a backup, just in case
//  power is lost while erasing one of those pages
struct sys_cfg code SYS_nvcfg0[(512/sizeof(struct sys_cfg))] _at_ 0x3a00;  /* 3a00 - 3bff */
struct sys_cfg code SYS_nvcfg1[(512/sizeof(struct sys_cfg))] _at_ 0x3c00;  /* 3c00 - 3dff */

// The ram copy of it
static struct sys_cfg SYS_cfg = {250.0, 2.0, 1.0, 0.0, 0, 1, 0};

// ISR variables
static volatile u8 delayFlag = 0;
static u8 beepTimeout = 0;
static u8 lastKeyState = 0;
static u8 readCtr = 0;
static u8 stableKeyState = 0;

static u16 cfgTimeToStore = 0;
static u8 execCtrlTimeout = 0;	// Execute control loop timer
static bit execControl = 0; 	// Execute control loop

static float targetWarning = 0.0;

// LCD ram
static u8 lcd_ram[9] = { 0,0,0,0,0,0,0,0,0};

//-----------------------------------------------------------------------------
// Initialization Subroutines
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// PORT_Init
//-----------------------------------------------------------------------------
//
// Configure the Crossbar and GPIO ports.
//
//	P0.1 O Buzzer
//	P0.2 O Solder On
//  P0.4 - UART TX
//  P0.5 - UART RX
//	P1.2 A ThermoCouple
//	P1.4 A ThermoCouple GND
//	P2.0 I PushButton (Active LO) - Unused
//	P2.1 I PushButton (Active LO) - Celcius/Farenheit
//	P2.2 I PushButton (Active LO) - Down
//	P2.3 I PushButton (Active LO) - Up
//	P3.4 IO HT DATA
//	P3.3 O  HT WR
//  P3.2 O  HT CS


static void PORT_Init (void)
{
	XBR0    = 0x01;                     // Enable UART on P0.4(RX) and P0.5(TX)              
	XBR1    = 0x41;                     // Enable crossbar and enable weak pull-ups, CEX0 routed to port pin

	P0MDIN  = 0xff;					   // P0 not configured as analog input
	P1MDIN  = 0xeb;					   // P1.2 and P1.4 configured as analog inputs
	P2MDIN  = 0xff;					   // P2 not configured as analog input
	P3MDIN  = 0xff;					   // P3 not configured as analog input

	P0MDOUT = 0x14;                     // enable UTX and SOLDERON as push-pull output
	P1MDOUT = 0x00;
	P2MDOUT = 0x00;
	P3MDOUT = 0x1c;                     // enable HT controls as pushpull outputs

	P0SKIP  = 0x03;                     // Skip P0.0 and P0.1 -> CEX0 will be assigned to P0.2
	P1SKIP  = 0x14;					   // Analog inputs skipped by crossbar
	P2SKIP  = 0x00;
}

//-----------------------------------------------------------------------------
// SYSCLK_Init
//-----------------------------------------------------------------------------
//
// This routine initializes the system clock to use the internal oscillator
// at its maximum frequency.
// Also enables the Missing Clock Detector.
//

static void SYSCLK_Init (void)
{
	OSCICN = 0x83;                      // Configure internal oscillator for
	// its maximum frequency
	CLKSEL = 0x00;					   // SYSCLK derived from internal oscilator
	RSTSRC = 0x04;                      // Enable missing clock detector
}

//-----------------------------------------------------------------------------
// PWM_Init
//-----------------------------------------------------------------------------
//
// This routine initializes the PCA as a PWM8 output on P0.2 using TMR0 as
//  timebase. Gives a PWM signal with 373hz periods
//
static void PWM_Init (void)
{
	// Configure Timer0 as a clock divider to lower down the pwm period
	TL0 = TH0 = 0x00;                  	// init Timer0
	TMOD &= ~0x0f;                      // TMOD: timer 0 in 8-bit autoreload
	TMOD |=  0x02;                       
	TR0 = 1;                            // START Timer0

	// configure the PCA
	PCA0MD = 0x04;						// disable CF interrupt
	// PCA time base = Tmr0 overflow
	PCA0CPL0 = 0xFF;					// initialize PCA PWM value			
	PCA0CPH0 = 0xFF;
	PCA0CPM0 = 0x42;					// CCM0 in 8-bit PWM mode
	PCA0CN = 0x40;						// enable PCA counter
}

//-----------------------------------------------------------------------------
// PWM_SetDutyCycle
//-----------------------------------------------------------------------------
//
// Updates the PWM value duty cycle (0-256) .. 0=off, 256=full on
//
// Parameters:
//  val=Duty cycle (0=off, 256=full on)
//
static void PWM_SetDutyCycle(u16 val)
{
	if (!val) {
		PCA0CPM0 &= 0xBF;				// Clear ECOM0 will hold output to 0.
		return;
	}

	// Otherwise, write the desired duty cycle
	PCA0CPH0 = (u8)(256U-val);
}

//-----------------------------------------------------------------------------
// ADC_Init ADBUSY, LP tracking, no Interrupt, ADC disabled
//-----------------------------------------------------------------------------
//
// Configure ADC0 to use ADBUSY as conversion source, and to sense the output 
// of the temp sensor.  Disables ADC end of conversion interrupt. Leaves ADC 
// disabled.
//
static void ADC_Init (void)
{
	ADC0CN = 0x00;                      // ADC0 disabled; tracking
	// mode; ADC0 conversions are initiated 
	// on a write to ADBusy
	AMX0P  = 0x1E;                      // Temp sensor selected at + input
	AMX0N  = 0x1F;                      // Single-ended mode

	ADC0CF = (SYSCLK/3000000) << 3;     // ADC conversion clock <= 3MHz  ,  right-justified 

	REF0CN = 0x0e;                      // enable temp sensor, VREF = VDD, bias
	// generator is on.

	EIE1 &= ~0x08;                      // Disable ADC0 EOC interrupt
}

//-----------------------------------------------------------------------------
// UART_Init
//-----------------------------------------------------------------------------
//
// Configure the UART0 using Timer1, for <BAUDRATE> and 8-N-1.
//
static void UART_Init (void)
{
	SCON0 = 0x10;                       // SCON0: 8-bit variable bit rate
	//        level of STOP bit is ignored
	//        RX enabled
	//        ninth bits are zeros
	//        clear RI0 and TI0 bits
	if (SYSCLK/BAUDRATE/2/256 < 1) {
		TH1 = -(SYSCLK/BAUDRATE/2);
		CKCON &= ~0x0B;                  // T1M = 1; SCA1:0 = xx
		CKCON |=  0x08;
	} else if (SYSCLK/BAUDRATE/2/256 < 4) {
		TH1 = -(SYSCLK/BAUDRATE/2/4);
		CKCON &= ~0x0B;                  // T1M = 0; SCA1:0 = 01                  
		CKCON |=  0x09;
	} else if (SYSCLK/BAUDRATE/2/256 < 12) {
		TH1 = -(SYSCLK/BAUDRATE/2/12);
		CKCON &= ~0x0B;                  // T1M = 0; SCA1:0 = 00
	} else {
		TH1 = -(SYSCLK/BAUDRATE/2/48);
		CKCON &= ~0x0B;                  // T1M = 0; SCA1:0 = 10
		CKCON |=  0x02;
	}

	TL1 = TH1;                          // init Timer1
	TMOD &= ~0xf0;                      // TMOD: timer 1 in 8-bit autoreload
	TMOD |=  0x20;                       
	TR1 = 1;                            // START Timer1
	TI0 = 1;                            // Indicate TX0 ready
}

//-----------------------------------------------------------------------------
// ISR_Init SYSCLK with Interrupt
//-----------------------------------------------------------------------------
//
// Configure Timer2 to auto-reload at interval specified by <counts> (with 
// interrupts generated) using SYSCLK as its time base.
//
static void ISR_Init (int counts)
{
	TMR2CN = 0x00;                      // STOP Timer2; Clear TF2H and TF2L;
	// disable low-byte interrupt; disable
	// split mode; select internal timebase
	CKCON  = 0x10;                      // Timer2 uses SYSCLK as its timebase

	TMR2RL  = -counts;                  // Init reload values
	TMR2    = -counts;                  // Init Timer2 with reload value
	ET2 = 1;                            // enable Timer2 interrupts
	TR2 = 1;                            // start Timer2
}

//-----------------------------------------------------------------------------
// ISRs
//-----------------------------------------------------------------------------
static void Tmr2_Isr (void) interrupt 5
{
	u8 keyState;
	TF2H = 0;                           // Clear Timer2 overflow flag

	// Apagar el beeper cuando se requiera
	if (beepTimeout) {
		beepTimeout--;
		if (!beepTimeout) {
			BUZZER = 1; // Turn off buzzer
		}
	}	

	// Mantener el timer para ejecucion del bucle de control
	if (!execControl) {
		if (execCtrlTimeout) {
			execCtrlTimeout--;
			if (!execCtrlTimeout) {
				execControl = 1;
			}
		}
	}

	//	P2.0 I PushButton (Active LO) - Unused
	//	P2.1 I PushButton (Active LO) - Celcius/Farenheit
	//	P2.2 I PushButton (Active LO) - Down
	//	P2.3 I PushButton (Active LO) - Up

	keyState = (P2 & 0xF) ^ 0xF;
	if (keyState != lastKeyState) {
		lastKeyState = keyState;
		readCtr = 30;
	} else {
		if (readCtr > 0) {
			readCtr--;
		} else {
			stableKeyState = lastKeyState;
		}
	}

	// Set the delay flag
	delayFlag = 1;
}

//-----------------------------------------------------------------------------
// DelayUs
//-----------------------------------------------------------------------------
//
// Performs a Delay in Us
//
static void DelayUs(u16 us)
{
	u16 delay = us;
	while (delay > 0) {
		_nop_();
		delay --;
	};
}

#if 0
//-----------------------------------------------------------------------------
// DelayMs
//-----------------------------------------------------------------------------
//
// Performs a Delay in Ms
//
static void DelayMs(u16 ms)
{
	unsigned int count;   
	for (count = ((TIMER2_RATE*ms + 999)/1000); count != 0; count--) {
		while (!delayFlag);                   // wait for overflow
		delayFlag = 0;                        // clear overflow indicator
	}
}
#endif

//------------------------------------------------------------------------------
// VDDMon_Init
//------------------------------------------------------------------------------
//
// Initializes the VDD monitor module to protect FLASH writes
//
static void VDDMon_Init (void)
{
	VDM0CN = 0x80;						// Turn Vdd monitor ON
	DelayUs(100);					    // Stabilization

	RSTSRC = 0x06;						// Vdd monitor and missing clk source resets
}

//-----------------------------------------------------------------------------
// ADC_ReadChannel
//-----------------------------------------------------------------------------
//
// This routine averages 16383 ADC samples from channel <channel> and returns 
// a 16-bit unsigned result.
// 

#define ADCCH_TEMPSENSOR	0	/* Internal sensor */
#define ADCCH_THERMO		1	/* Thermocouple */

static u16 ADC_ReadChannel(u8 channel)
{
	u16 i;                       // Sample counter
	u32 accumulator=0L;       	// Here's where we integrate the
	// ADC samples
	u16 currval;

	if (channel == ADCCH_TEMPSENSOR) {
		AMX0P  = 0x1E;                      // Temp sensor selected at + input
		AMX0N  = 0x1F;                      // Single-ended mode
	} else {
		//	P1.2 A ThermoCouple
		//	P1.4 A ThermoCouple GND - Do not use...
		AMX0P  = 0x02;                      // + at P1.2
		AMX0N  = 0x1F;                      // Single ended mode
	}

	AD0EN = 1;                          // Enable ADC0

	DelayUs(10);

	AD0INT = 0;
	AD0BUSY = 1;

	// read the ADC value and add to running total
	i = 0;
	do 
	{
		while (!AD0INT);                 // Wait for conversion to complete
		AD0INT = 0;                      // Clear end-of-conversion indicator

		currval=ADC0;                    // Store latest ADC conversion          
		AD0BUSY = 1;                     // Initiate conversion
		accumulator += currval;          // Accumulate
		i++;                             // Update counter
	} while (i != 16383);

	AD0EN = 0;                          // Disable ADC0

	// shift to obtain a 16-bit result (14 + 10 = 24 - 8 = 16) bits
	return (u16) (accumulator >> 8);
}

//-----------------------------------------------------------------------------
// KBD_ReadKey
//-----------------------------------------------------------------------------
//
// Reads a key until no bounces left.
//
#define KBD_CFTOGGLE 0x2
#define KBD_DOWN	 0x4
#define KBD_UP		 0x8
static u8 KBD_ReadKey(void)
{
	return stableKeyState;
}

/*------------------------------------------------------------------------------*/

//------------------------------------------------------------------------
// HT1621 commands 
//------------------------------------------------------------------------

#define HTCMD_BIAS   	0x52   /*010abXcX = 4 commons */
#define HTCMD_SYSEN  	0x02   
#define HTCMD_LCDON  	0x06   
#define HTCMD_LCDOFF 	0x04   
#define HTCMD_XTAL   	0x30
#define HTCMD_TONEOFF  	0x10
#define HTCMD_TONEON	0x12
#define HTCMD_TONE2K	0xc0
#define HTCMD_TONE4K	0x80

//-----------------------------------------------------------------------------
// HT_SendBit
//-----------------------------------------------------------------------------
//
// Send <cnt> bits of the <dat> number, from msbit to lsbit to the HT1621
//
static void HT_SendBit(u8 dat, u8 cnt)
{   
	u8 i=0;   
	for(i=0; i<cnt; i++)   
	{   
		if(dat&0x80)    
			HTDAT = 1;   
		else   
			HTDAT = 0;   
		_nop_();_nop_();_nop_();_nop_();   
		HTWR = 0;   
		_nop_();_nop_();_nop_();_nop_();   
		HTWR = 1;   
		_nop_();_nop_();_nop_();_nop_();   
		dat <<= 1;   
	}   
}   

//-----------------------------------------------------------------------------
// HT_SendDataBit
//-----------------------------------------------------------------------------
//
// Send <cnt> bits of the <dat> number, from lsbit to msbit to the HT1621
//
static void HT_SendDataBit(u8 dat, u8 cnt)
{   
	u8 i=0;   
	for(i=0; i<cnt; i++)   
	{   
		if(dat&0x01)   
			HTDAT = 1;   
		else    
			HTDAT = 0;   
		_nop_();_nop_();_nop_();_nop_();   
		HTWR = 0;   
		_nop_();_nop_();_nop_();_nop_();   
		HTWR = 1;   
		_nop_();_nop_();_nop_();_nop_();   
		dat >>= 1;   
	}   
}   

static void HT_SendCmd(u8 command)   
{   
	HTCS = 0;   
	HT_SendBit(0x80, 4);   
	HT_SendBit(command, 8);   
	HTCS = 1;   
}   

/* 4bit succesive writes*/
static void WriteAll_1621(u8 addr, u8 *p, u8 cnt)   
{   
	u8 i=0;   
	HTCS = 0;   
	HT_SendBit(0xA0, 3);   /*write*/
	HT_SendBit(addr, 6);   
	for(i=0; i<cnt; i++,p++)   
	{   
		HT_SendDataBit(*p, 8);    
	}   
	HTCS = 1;   
}   

//-----------------------------------------------------------------------------
// HT_Init
//-----------------------------------------------------------------------------
//
// Initialize and turn on the LCD of the solder station
//
static void HT_Init()   
{   
	HT_SendCmd(HTCMD_XTAL);   
	HT_SendCmd(HTCMD_BIAS);   
	HT_SendCmd(HTCMD_SYSEN);   
	HT_SendCmd(HTCMD_LCDON);   
}   

//-----------------------------------------------------------------------------
// BEEP_Do
//-----------------------------------------------------------------------------
//
// Does a Beep
//
static void BEEP_Do(void)
{
	BUZZER = 0; 		// Turn on buffer
	beepTimeout = 30;	// 300 ms
}

//-----------------------------------------------------------------------------
// LCD_SetBit
//-----------------------------------------------------------------------------
//
//  Sets or clears, depending on <val> value a bit <seg> of the LCD controller memory
//
static u8 LCD_SetBit(u8 seg,u8 val)
{
	u8 nval = val ? 1 : 0;
	u8 mask = 1<<(seg&7);
	u8 pos = seg >> 3;
	u8 cval = (lcd_ram[pos] & mask) ? 1 : 0;
	if (nval == cval) 
		return 0;
	if (val)
		lcd_ram[pos] |= mask;
	else
		lcd_ram[pos] &= ~mask;
	return 1; 
}


//-----------------------------------------------------------------------------
// LCD_SetDigit
//-----------------------------------------------------------------------------
//
//  Sets a digit <pos> from the main or auxiliary displays, as specified by <main>
// to the given number <nr>
//
static u8 LCD_SetDigit(u8 main, u8 pos, u8 nr)
{
	// Display Pattern for each digit
	//  b
	//  -
	//a| |c
	// +-+g
	//f| |d
	//  -
	//  e
	// gfedcba
	//


	//
	// TEMP  +-a-+ +---+ +---+   
	//       | A b | B | | C |		  HEATON
	//       f   | |   | |   | o +--+
	//       +g--+ +---+ +---+   |D |  WAIT
	//       |   c |   | |   |	 +--+
	//       e   | |   | |   |   |  |  ERROR
	//       +-d-+ +---+ +---+   +--+ 
	//
	// SET   +---+ +---+ +---+   
	//       | E | | F | | G |
	//       |   | |   | |   | o +--+
	//       +---+ +---+ +---+   |H | 
	//       |   | |   | |   |   +--+
	//       |   | |   | |   |   |  |
	//       +---+ +---+ +---+   +--+ 
	//
	//
	//0 H-a
	//1 H-b
	//2 H-c
	//3 H-d
	//4 ---
	//5 ERROR
	//6 WAIT
	//7 --
	//8 --
	//9 --
	//10 --
	//11 HEATON
	//12 D-d
	//13 D-c
	//14 D-b
	//15 D-a
	//16 D-e
	//17 D-g
	//18 D-f
	//19 o
	//20 C-d
	//21 C-c
	//22 C-b
	//23 C-a
	//24 C-e
	//25 C-g
	//26 C-f
	//27 ---
	//28 B-d
	//29 B-c
	//30 B-b
	//31 B-a
	//32 B-e
	//33 B-g
	//34 B-f
	//35 ---
	//36 A-d
	//37 A-c
	//38 A-b
	//39 A-a
	//40 A-e
	//41 A-g
	//42 A-f
	//43 TEMP
	//44 SET
	//45 E-f
	//46 E-g
	//47 E-e
	//48 E-a
	//49 E-b
	//50 E-c
	//51 E-d
	//52 ---
	//53 F-f
	//54 F-g
	//55 F-e
	//56 F-a
	//57 F-b
	//58 F-c
	//59 F-d
	//60 ---
	//61 G-f
	//62 G-g
	//63 G-e
	//64 G-a
	//65 G-b
	//66 G-c
	//67 G-d
	//68 o (low)
	//69 H-f
	//70 H-g
	//71 H-e

	//
	//

#define SEG_A 0x01
#define SEG_B 0x02
#define SEG_C 0x04
#define SEG_D 0x08
#define SEG_E 0x10
#define SEG_F 0x20
#define SEG_G 0x40

	// Pattern for each digit
	static const u8 pattern[] = {
		SEG_A|SEG_B|SEG_C|SEG_D|SEG_E|SEG_F,		// 0
		SEG_C|SEG_D,								// 1
		SEG_B|SEG_C|SEG_E|SEG_F|SEG_G,				// 2
		SEG_B|SEG_C|SEG_D|SEG_E|SEG_G,				// 3
		SEG_A|SEG_C|SEG_D|SEG_G,					// 4
		SEG_A|SEG_B|SEG_D|SEG_E|SEG_G,				// 5
		SEG_A|SEG_B|SEG_D|SEG_E|SEG_F|SEG_G,		// 6
		SEG_B|SEG_C|SEG_D,					        // 7
		SEG_A|SEG_B|SEG_C|SEG_D|SEG_E|SEG_F|SEG_G,	// 8
		SEG_A|SEG_B|SEG_C|SEG_D|SEG_E|SEG_G,		// 9
		SEG_A|SEG_B|SEG_C|SEG_D|SEG_F|SEG_G,		// A
		SEG_A|SEG_D|SEG_E|SEG_F|SEG_G,				// b
		SEG_A|SEG_B|SEG_E|SEG_F,					// C
		SEG_C|SEG_D|SEG_E|SEG_F|SEG_G,				// d
		SEG_A|SEG_B|SEG_E|SEG_F|SEG_G,				// E
		SEG_A|SEG_B|SEG_F|SEG_G						// F
	};

	// Segments for each number 
	static const u8 segment[2][4][7] = {
		// High 
		{
			{ // D
				18,15,14,13,12,16,17
			},
			{ // C
				26,23,22,21,20,24,25
				},
				{ // B
					34,31,30,29,28,32,33
				},
				{ // A
					42,39,38,37,36,40,41
					}
		},
			// Low
		{
			{ // H
				69,0,1,2,3,71,70
			},
			{ // G
				61,64,65,66,67,63,62
				},
				{ // F
					53,56,57,58,59,55,54
				},
				{ // E
					45,48,49,50,51,47,46
					}
		}
	};

	u8 shouldupdate = 0;
	u8 patt = pattern[nr];
	u8 spos, smask;
	for (spos = 0, smask = 1; spos < 7; spos++, smask <<= 1) {
		shouldupdate |= LCD_SetBit(segment[main][pos][spos],patt & smask);
	} 
	return shouldupdate;
}

//-----------------------------------------------------------------------------
// LCD_SetNbr
//-----------------------------------------------------------------------------
//
//  Sets a number <nbr> in one of the LCD displays (main or auxiliary) as
//  specified by <main>
//
// 
static void LCD_SetNbr(u16 nbr, u8 main)
{
	// Print the number into the display
	u8 shouldupdate = 0;
	u8 pos;
	u16 n = nbr;
	for (pos = 1; pos < 4; pos++) {
		u8 dig = n % 10;
		shouldupdate |= LCD_SetDigit(main,pos,dig);
		n = n / 10;
	}

	// Update LCD
	if (shouldupdate)
		WriteAll_1621(0, lcd_ram, 9);
}

#define LCDICON_ERROR 	5
#define LCDICON_WAIT  	6
#define LCDICON_HEATON 11
#define LCDICON_DEG_H  19
#define LCDICON_TEMP   43
#define LCDICON_SET    44
#define LCDICON_DEG_L  68

//-----------------------------------------------------------------------------
// LCD_SetIcon
//-----------------------------------------------------------------------------
//
//  Show or hide, as specified by <state> the icon specified by <icon>
// 
static void LCD_SetIcon(u8 icon, u8 state)
{
	if (LCD_SetBit(icon,state)) {
		// Update LCD
		WriteAll_1621(0, lcd_ram, 9);
	}
}

//-----------------------------------------------------------------------------
// LCD_SetTempSymbol
//-----------------------------------------------------------------------------
//
//  Show the "C" or "F" letters in the display, as specified by <farh>
// 

static void LCD_SetTempSymbol(u8 farh)
{
	u8 v = farh ? 0xF : 0xC;
	u8 shouldupdate = LCD_SetDigit(0,0,v);
	shouldupdate |= LCD_SetDigit(1,0,v);

	// Update LCD
	if (shouldupdate) 
		WriteAll_1621(0, lcd_ram, 9);
}

//-----------------------------------------------------------------------------
// FLASH_Erase
//-----------------------------------------------------------------------------
//
//  Erases the 512 byte page pointed by codePtr
static void FLASH_Erase(void code * codePtr)
{
	bit EA_state=EA;                    // Preserves EA state				
	EA = 0;                             // Disable interrupts

	// Erase flash...
	PSCTL |= 0x02;					    // Erase enable
	FLKEY=0xA5;                         // Input first key code
	FLKEY=0xF1;                         // Input second key code,
	PSCTL |= 0x01;                      // Enable FLASH Writes
	*((u8 xdata *)codePtr) = 0xFF;	    // Erase flash
	PSCTL = 0x00;                       // disable FLASH Writes and erases
	EA = EA_state;                      // Restore interrupt state
}

//-----------------------------------------------------------------------------
// FLASH_Write
//-----------------------------------------------------------------------------
//
//  Write to the given address in flash
static void FLASH_Write(void code * codePtr,void* cfgPtr,u8 sz)
{
	bit EA_state=EA;                    // Preserves EA state				
	EA = 0;                             // Disable interrupts

	// Write to flash...
	do {
		FLKEY=0xA5;                     // Input first key code
		FLKEY=0xF1;                     // Input second key code,
		// FLASH is now unlocked

		PSCTL |= 0x01;                  // Enable FLASH Writes
		*((u8 xdata *)codePtr) = *((u8*)cfgPtr);   		   // Write high byte of SYS_cfg.setpoint

		PSCTL &= ~0x01;                 // disable FLASH Writes

		codePtr = ((u8 xdata *)codePtr) + 1;                      // Move to next
		cfgPtr = ((u8*)cfgPtr) + 1;
	} while (--sz);

	PSCTL = 0x00;                       // Disable FLASH Writes
	EA = EA_state;                      // Restore interrupt state
}

//-----------------------------------------------------------------------------
// FLASH_NeedErase
//-----------------------------------------------------------------------------
//
//  Check if we need to erase a page to write to it
static u8 FLASH_NeedErase(void code * codePtr,void* cfgPtr,u8 sz)
{
	u8 ret = 0;

	while (sz--) {
		u8 ramByte   = *((u8*)cfgPtr);
		u8 flashByte = *((u8 code *)codePtr);

		// Compute changed bits.
		u8 chgBits = flashByte ^ ramByte;

		// Compute bits that change from 0 to 1
		u8 chg0to1 = ramByte & chgBits;

		// If there is a need to change from 0 to 1, we need to erase
		if (chg0to1 != 0) {
			return 1; // Change needed
		}

		// Compare the next bytes
		codePtr = ((u8 code *)codePtr) + 1;                      // Move to next
		cfgPtr = ((u8*)cfgPtr) + 1;

	}

	// No changes needed
	return 0;
}

//-----------------------------------------------------------------------------
// FLASH_Verify
//-----------------------------------------------------------------------------
//
//  Check if written data was properly written
static u8 FLASH_Verify(void code * codePtr,void* cfgPtr,u8 sz)
{
	while (sz--) {
		// Compute changed bits.
		u8 c = *((u8 code *)codePtr);
		u8 r = *((u8*)cfgPtr);
		if (c != r)
			return 0;

		// Compare the next bytes
		codePtr = ((u8 code *)codePtr) + 1;                      // Move to next
		cfgPtr = ((u8*)cfgPtr) + 1;
	}

	// Ok
	return 1;
}

//-----------------------------------------------------------------------------
// Support Subroutines
//-----------------------------------------------------------------------------

//  Compute checksum for a SYS record
static u8 SC_cksum(struct sys_cfg code * sc)
{
	u8 cksum = 0, pos = 0;
	for (pos = 0; pos < sizeof(struct sys_cfg) - 1; pos++) {
		cksum += ((u8 code*)sc)[pos];
	}
	return cksum;
}

//-----------------------------------------------------------------------------
// SYSCFG_Init
//-----------------------------------------------------------------------------
//
// Read config from non volatile storage. Returns if config record health is bad (means
//  we need to store a new copy of it)
//
u8 SYSCFG_Init(void)
{
	u8 badhealth = 0; // No bad health for sysconfig records
	u8 pos = sizeof(SYS_nvcfg0) / sizeof(SYS_nvcfg0[0]); 

	// Scan for a valid configuration...
	while (pos) {
		pos--;

		// Check the master record ...
		if (SYS_nvcfg0[pos].invalid == 0 && SC_cksum(&SYS_nvcfg0[pos]) == SYS_nvcfg0[pos].cksum) {

			// Read config from nv storage
			SYS_cfg.setpoint = SYS_nvcfg0[pos].setpoint;
			SYS_cfg.use_fahrenheit = SYS_nvcfg0[pos].use_fahrenheit;
			SYS_cfg.pid = SYS_nvcfg0[pos].pid;
			badhealth = SYS_nvcfg1[pos].invalid != 0 || SC_cksum(&SYS_nvcfg1[pos]) != SYS_nvcfg1[pos].cksum;
			pos = 0xFF;
			break;
		} else 
		// Check the backup record ...
		if (SYS_nvcfg1[pos].invalid == 0 && SC_cksum(&SYS_nvcfg1[pos]) == SYS_nvcfg1[pos].cksum) {

			// Read config from nv storage
			SYS_cfg.setpoint = SYS_nvcfg1[pos].setpoint;
			SYS_cfg.use_fahrenheit = SYS_nvcfg1[pos].use_fahrenheit;
			SYS_cfg.pid = SYS_nvcfg1[pos].pid;
			pos = 0xFF;
			badhealth = 1; // BAD healt, as we are using the backup record
			break;
		}
	}

	// If no valid configuration, use a standard one
	if (pos != 0xFF) {
		SYS_cfg.setpoint = 240.0;
		SYS_cfg.use_fahrenheit = 0;
		
		SYS_cfg.pid.Kp = 9.5; // 4-2-1
		SYS_cfg.pid.Ki = 0.01;
		SYS_cfg.pid.Kd = 0;

		//SYS_cfg.pid.Kp = 3.500;                                 //Initial P value
		//SYS_cfg.pid.Ki = 1.875;                                 //Initial I value
		//SYS_cfg.pid.Kd = 3.250;                                 //Initial D value 
		
//
//  5% -> 212 deg
// 10% -> 242 deg
// 15% -> 354 deg
// 20% -> 407 deg
//

	}

	// Estimate warning as a percent (10%) of setpoint
	targetWarning = SYS_cfg.setpoint * 0.1;

	// Bad health
	return badhealth;
}

//-----------------------------------------------------------------------------
// SYSCFG_Store
//-----------------------------------------------------------------------------
//
// Store config to non volatile storage
//
void SYSCFG_Store(void)
{
	u8 cksum = 0, pos = 0;
	
	// Config is valid
	SYS_cfg.invalid = 0;
	
	// Compute checksum
	for (pos = 0; pos < sizeof(struct sys_cfg) - 1; pos++) {
		cksum += ((u8*)&SYS_cfg)[pos];
	}
	SYS_cfg.cksum = cksum;

	while (1) {
		s8 pos;

		// Scan for a valid configuration...
		for (pos = sizeof(SYS_nvcfg0) / sizeof(SYS_nvcfg0[0]) - 1; pos >= 0; pos--) {
			if (SYS_nvcfg0[pos].invalid == 0 || SYS_nvcfg1[pos].invalid == 0) {
				break;
			}
		}

		// No valid config found, erase flash page, and write the config there
		if (pos < 0) {
			pos = 0;
			FLASH_Erase(&SYS_nvcfg0[pos]);
			FLASH_Write(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg));
			FLASH_Erase(&SYS_nvcfg1[pos]);
			FLASH_Write(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg));
		} else 

		// If no transitions from 0 to 1, just reuse the current slot.
		if (!FLASH_NeedErase(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg)) && 
			!FLASH_NeedErase(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg))) { 

			// Store the config into the new slot
			FLASH_Write(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg));
			FLASH_Write(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg));
		} else {

			// Need a transition...
			pos++;
			if (pos >= sizeof(SYS_nvcfg0) / sizeof(SYS_nvcfg0[0])) {
			 	// No more slots free, erase page and prepare to store results in first slot
				pos = 0;
				FLASH_Erase(&SYS_nvcfg0[pos]);
				FLASH_Write(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg));
				FLASH_Erase(&SYS_nvcfg1[pos]);
				FLASH_Write(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg));
			} else {
				// Some slots free, just use them
				FLASH_Write(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg));
				FLASH_Write(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg));
			}
		}

		// Verify both
		if (FLASH_Verify(&SYS_nvcfg0[pos],&SYS_cfg,sizeof(SYS_cfg)) && 
			FLASH_Verify(&SYS_nvcfg1[pos],&SYS_cfg,sizeof(SYS_cfg)))
			break;

	};
} 

//-----------------------------------------------------------------------------
// Temperature Sensor Calibration PARAMETERS
//-----------------------------------------------------------------------------
#define VREF               3300UL        // ADC Voltage Reference (mV)

#if 0
//-----------------------------------------------------------------------------
// SENSOR_GetAmbTemp()
//-----------------------------------------------------------------------------
//
// Returns ambient temperature in celcius in 11.5 format
static float SENSOR_GetAmbTemp (void)
{
	float result = (float) ADC_ReadChannel(ADCCH_TEMPSENSOR);
	result -= 17813.0; 	// (897UL * 65536UL) / VREF;	   // 17813
	result *= 0.01503108234068006840316768317297; // (65536UL * 335UL) / (100UL * VREF);
	return result;
}
#endif

//-----------------------------------------------------------------------------
// SENSOR_GetSolderTemp()
//-----------------------------------------------------------------------------
//
// Returns solder temperature in celcius. The sensor gives 0.00125 V/C or
//  100 steps per degree.
// 
static float SENSOR_GetSolderTemp(void)
{
	return ((float)ADC_ReadChannel(ADCCH_THERMO)) * 0.01007081038552321007081038552321;
}

//-----------------------------------------------------------------------------
// Celcius2Fahrenheit()
//-----------------------------------------------------------------------------
//
// Converts Celcius to Fahrenheit
static float Celcius2Fahrenheit(float Tc)
{
	return (Tc + 32.0) * (9.0/5.0);
}

//-----------------------------------------------------------------------------
// Fahrenheit2Celcius()
//-----------------------------------------------------------------------------
//
// Converts Fahrenheit to Celcius
static float Fahrenheit2Celcius(float Tf)
{
	return ( Tf * (5.0/9.0)) - 32.0;
}

//-----------------------------------------------------------------------------
// ControlLoop
//-----------------------------------------------------------------------------
//
// Executes the PID loop that controls temperature of the soldering tip, or the calibration, 
// if needed.
//

//#define __LOG_RESPONSE 1
#ifdef __LOG_RESPONSE
u16 tempLog[400];
u16 tempLogIdx = 0;
u16 timeDivider = 3000; // 5 mins for stabilization purposes
#endif

static void ControlLoop(void)
{
	float Temp;
	u16 v;

	// Check if we have to execute the control loop...
	if (!execControl)
		return;

	execCtrlTimeout = 10;	// Execute control loop 10 times per second
	execControl = 0; 		// Delay execution until timer decremented to 0

	// Read solder temperature in celsius
	Temp = SENSOR_GetSolderTemp();

#ifdef __LOG_RESPONSE
	timeDivider--;
	if (!timeDivider) {
		timeDivider = 10;
		if (tempLogIdx < 400) {
			u16 t = ((u16)Temp); // Temperature
			tempLog[tempLogIdx++] = t;
		} else {
			PWM_SetDutyCycle(0);
			tempLogIdx--;
		}
	}
#endif	
	
	// Flag error if temperature too high
	if (Temp > 600.0) {

		// Set default icons
		LCD_SetIcon(LCDICON_DEG_H,1);
		LCD_SetIcon(LCDICON_SET,1);
		LCD_SetIcon(LCDICON_DEG_L,1);

		// Flag error
		LCD_SetIcon(LCDICON_ERROR, 1);

		// Flag wait 
		LCD_SetIcon(LCDICON_WAIT, 1);

		// Turn off solder
		PWM_SetDutyCycle(0);
		LCD_SetIcon(LCDICON_HEATON, 0);

		// Clean display if error.

		// Display setpoint - Setpoint is always stored in the selected unit
		LCD_SetNbr((u16)SYS_cfg.setpoint,1);
	
		// Set the icon
		LCD_SetTempSymbol(SYS_cfg.use_fahrenheit);

		// Display temperatures [main]
		LCD_SetNbr(0, 0);

	} else {
				
		// Default PID temperature control

		// Remove error
		LCD_SetIcon(LCDICON_ERROR, 0);

#ifdef __LOG_RESPONSE		
		v = timeDivider > 10 ? (255UL * 15UL / 100UL) : (255UL * 20UL / 100UL); // 10% to 20% time response
#else
		// Calculate PID controller.
		v = (u16) PID_Compute(&PID_State, Temp, SYS_cfg.setpoint);
#endif
		
		
		// Use a PWM value proportional to the difference and the error
		PWM_SetDutyCycle(v);

		if (v) {
				// Set the Icon
				LCD_SetIcon(LCDICON_HEATON, 1);
		} else {
				// Clear the Icon
				LCD_SetIcon(LCDICON_HEATON, 0);
		}

		// Display setpoint - Setpoint is always stored in the selected unit
		LCD_SetNbr((u16)SYS_cfg.setpoint,1);
	
		// Set the icon
		LCD_SetTempSymbol(SYS_cfg.use_fahrenheit);

		// If using fahrenheit, convert measured temp to it
		if (SYS_cfg.use_fahrenheit) {
			Temp = Celcius2Fahrenheit(Temp);
		}

		// Flag wait if temperature low or high if we are not in autocalibration mode
		LCD_SetIcon(LCDICON_WAIT, ((Temp < (SYS_cfg.setpoint - targetWarning)) || 
									 (Temp > (SYS_cfg.setpoint + targetWarning))) 
										? 1 : 0);

		// Display temperatures [main]
		LCD_SetNbr((u16)Temp, 0);
	}

	// Store config if dirty

	// Keep the config store timeout
	if (cfgTimeToStore) {
			cfgTimeToStore --;
			if (!cfgTimeToStore) {
				// Store configuration
				SYSCFG_Store();
			}
	}
}

#if 0
//-----------------------------------------------------------------------------
// KBD_TimedWaitKey
//-----------------------------------------------------------------------------
//
// Wait until a key is pressed with timeout in <ms>
//
static u8 KBD_TimedWaitKey(u16 ms)
{
	unsigned int count; 
	u8 key = 0;  
	for (count = ((TIMER2_RATE*ms + 999)/1000); count != 0; count--) {
		while (!delayFlag) {
			// A key was detected, break loop
			if ((key = KBD_ReadKey()) != 0)
				break;
			// Execute the control loop while waiting
			ControlLoop(); 
		};
		delayFlag = 0;                        // clear overflow indicator
	}
	return key;
}
#endif

//-----------------------------------------------------------------------------
// KBD_TimedReleaseWait
//-----------------------------------------------------------------------------
//
// Wait until all keys are released with timeout in <ms>
//
static u8 KBD_TimedReleaseWait(u16 ms)
{
	unsigned int count; 
	u8 key = 0;  
	for (count = ((TIMER2_RATE*ms + 999)/1000); count != 0; count--) {
		while (!delayFlag) {
			// All keys were released, break loop
			if ((key = KBD_ReadKey()) == 0)
				break;
			// Execute the control loop while waiting
			ControlLoop(); 
		};
		delayFlag = 0;                        // clear overflow indicator
	}
	return key;
}

//-----------------------------------------------------------------------------
// KBD_WaitKey
//-----------------------------------------------------------------------------
//
// Wait until a key is pressed 
//
static u8 KBD_WaitKey(void)
{
	u8 key = 0;  
	while (1) {
		// A key was detected, break loop
		if ((key = KBD_ReadKey()) != 0)
			break;
		// Execute the control loop while waiting
		ControlLoop(); 
	};
	return key;
}

//-----------------------------------------------------------------------------
// KBD_ReleaseWait
//-----------------------------------------------------------------------------
//
// Wait until all keys are released
//
static void KBD_ReleaseWait(void)
{
	u8 key = 0;  
	while (1) {
		// All keys were released, break loop
		if ((key = KBD_ReadKey()) == 0)
			break;
		// Execute the control loop wile waiting
		ControlLoop(); 
	};
}

// Keyboard acceleration table
static const struct {
	u16   delay;	// Delay to execute next increment
	u8    steps;	// Consecutive steps at this level
	float delta;	// Delta advance at this level
} code acc_tab[4] = {
	{1000,  2, 1.0}, // 2 secs,  1/s
	{ 500,  4, 3.0}, // 2 secs,  6/s
	{ 250,  8, 8.0}, // 2 secs, 32/s
	{ 100,255,16.0}  // -     ,160/s 
};


//-----------------------------------------------------------------------------
// Main program
//-----------------------------------------------------------------------------
//
void main (void) 
{
	u8 pos = 0;
	u8 key;
	u16 step;
	u16 stepidx;


	// Disable Watchdog timer
	PCA0MD &= ~0x40;                    	// WDTE = 0 (clear watchdog timer enable)
	PORT_Init();                        	// Initialize Port I/O
	SYSCLK_Init();                     		// Initialize Oscillator
	VDDMon_Init(); 							// Vdd monitor reset
	ADC_Init();                       		// Init ADC0
	PWM_Init();
	ISR_Init(SYSCLK/TIMER2_RATE);    		// Init Timer 2 

	UART_Init();					   		// Init UART
	HT_Init();						   		// Init LCD controller
	SYSCFG_Init();					   		// Read System configuration
	
	PID_Create(&PID_State, SYS_cfg.pid.Kp, SYS_cfg.pid.Ki, SYS_cfg.pid.Kd, 50, DIRECT);

	// We are ready, enable ints
	EA = 1;
	execControl = 1;						// Execute control loop

#if 0
	// ... LCD segment test ...
	lcd_ram[0] = 1;
	do {

		// Update LCD
		WriteAll_1621(0, lcd_ram, 9);

		// Wait until key released
		while (KBD_ReadKey() != 0);

		// Wait until something pressed
		while ((key=KBD_ReadKey()) == 0);

		// Small beep
		BEEP_Do();

		//if (key == (KBD_CFTOGGLE|KBD_DOWN|KBD_UP))
		//	break;

		// Walk a bit ..
		lcd_ram[pos] <<= 1;
		if (lcd_ram[pos] == 0) {
			pos++;
			if (pos > 9)
				pos = 0;
			lcd_ram[pos] = 1;
		}

	} while (1);
#endif

	// Set default icons
	LCD_SetIcon(LCDICON_DEG_H,1);
	LCD_SetIcon(LCDICON_TEMP,1);
	LCD_SetIcon(LCDICON_SET,1);
	LCD_SetIcon(LCDICON_DEG_L,1);

	// Do the Ready beep
	BEEP_Do();

#if 0
	// ... LCD number display test ...
	{
		u16 n = 0;
		while (1) {
			LCD_SetNbr(n,0);
			LCD_SetNbr(n,1);
			DelayMs(20);
			n++;
			if (n > 999) n = 0;
		}
	}
#endif

	while (1) {

		// Wait until something pressed
		key = KBD_WaitKey();

		// Based on key, do...
		switch(key) {
		
		// Celsius - Farenheit toggle
		case KBD_CFTOGGLE:


			BEEP_Do();

			// Toggle units...
			SYS_cfg.use_fahrenheit ^= 1;

			// And convert them
			if (SYS_cfg.use_fahrenheit)
				SYS_cfg.setpoint = Celcius2Fahrenheit(SYS_cfg.setpoint);
			else
				SYS_cfg.setpoint = Fahrenheit2Celcius(SYS_cfg.setpoint);

			// Clamp values to displayable ones
			if (SYS_cfg.setpoint < 0.0)
				SYS_cfg.setpoint = 0.0;
			else if (SYS_cfg.setpoint > 999.9)
				SYS_cfg.setpoint = 999.9;

			// Estimate warning as a percent (10%) of setpoint
			targetWarning = SYS_cfg.setpoint * 0.1;

			// Force config flush with timeout so we don't destroy flash...
			cfgTimeToStore = 500;

			break;

		// Setpoint down with acceleration
		case KBD_DOWN:

			step = 0;
			stepidx = acc_tab[step].steps;
			do {
				BEEP_Do();

				SYS_cfg.setpoint -= acc_tab[step].delta;
				if (SYS_cfg.setpoint < 1.0) {
					SYS_cfg.setpoint = 1.0;
				}
				// Estimate warning as a percent (10%) of setpoint
				targetWarning = SYS_cfg.setpoint * 0.1;

				// Force config flush with timeout so we don't destroy flash...
				cfgTimeToStore = 500;

				stepidx--;
				if (!stepidx) {
					if (step < 3)
						step++;
					stepidx = acc_tab[step].steps;
				}
			} while (KBD_TimedReleaseWait(acc_tab[step].delay));

			break;

		// Setpoint up with acceleration
		case KBD_UP:

			step = 0;
			stepidx = acc_tab[step].steps;
			do {
				BEEP_Do();
				SYS_cfg.setpoint += acc_tab[step].delta;
				if (SYS_cfg.setpoint > 999.0) {
					SYS_cfg.setpoint = 999.0;
				}

				// Estimate warning as a percent (10%) of setpoint
				targetWarning = SYS_cfg.setpoint * 0.1;

				// Force config flush with timeout so we don't destroy flash...
				cfgTimeToStore = 500;

				stepidx--;
				if (!stepidx) {
					if (step < 3)
						step++;
					stepidx = acc_tab[step].steps;
				}
			} while (KBD_TimedReleaseWait(acc_tab[step].delay));

			break;
		}

		// Wait until key released
		KBD_ReleaseWait();
	}
}



